open EdisonPervasives
open StreamHeader

type 'a seq = 'a stream

let rec count_forced = function
  | Nil -> 0
  | Cons (_, tail) ->
      match !tail with
	| Lazy.Value _ -> 1 + count_forced (Lazy.force tail)
	| _-> 1

let unit = function x -> Cons (x, lazy Nil)
    
let rec map f = function
  | Nil -> Nil
  | Cons (head, tail) -> Cons (f head, lazy (map f (force tail)))

let rec plus = fun a b ->
  match a with
    | Nil -> b
    | Cons (head, tail) -> Cons (head, lazy (plus (force tail) b))

let zero () = Nil

let rec join = function
  | Nil -> Nil
  | Cons (head, tail) -> plus head (join (force tail))

let bind = fun stream f -> join (map f stream)

let rec make = fun n x ->
  match n with
    | 0 -> Nil
    | _ -> Cons (x, lazy (make (n - 1) x))

let rec get = fun stream n ->
  match (stream, n) with
    | (Cons (x, _), 0) -> x
    | (Cons (_, tail), _) -> get (force tail) (n - 1)
    | _ -> raise Empty

let rec set = fun stream n x ->
  match stream with
    | Nil when n = 0 -> Cons (x, lazy Nil)
    | Cons (_, tail) when n = 0 -> Cons (x, tail)
    | Cons (y, tail) -> Cons (y, lazy (set (force tail) (n - 1) x))
    | _ -> raise Empty

let compare = fun a b -> 0

(* From_list : les clotures sont directement forces 
 * puisque tous les lments de la liste sont connus 
 *)

let rec from_list = function
  | [] -> Nil
  | head :: tail -> Cons (head, ref (Lazy.Value (from_list tail)))

let rec to_list = function
  | Nil -> []
  | Cons (x, tail) -> x :: to_list (force tail)

let rec to_list_append a = function
  | Nil -> a
  | Cons (x, tail) -> x :: to_list_append a (force tail)

let rec reverse_onto result = function
  | Nil -> result
  | Cons (head, tail) -> 
      reverse_onto (Cons (head, ref (Lazy.Value result))) (force tail)

let rec reverse_append = fun stream result ->
  match stream with
    | Nil -> result
    | Cons (head, tail) -> 
	reverse_append (Cons (head, ref (Lazy.Value result))) (force tail)

let reverse = function stream -> reverse_onto Nil stream

let rec append_reverse_rec result = function
  | (Nil, stream) -> reverse_onto result stream
  | (stream, Nil) -> stream ++ result
  | (Cons (x, xs), Cons (y, ys)) ->
      Cons (x, lazy (append_reverse_rec (Cons (y, lazy result)) 
		       (force xs, force ys)))

let rec append_reverse = fun a b -> append_reverse_rec Nil (a, b)

let rec reverse_iter f = function
  | Nil -> ()
  | Cons (head, tail) -> reverse_iter f (force tail) ; f head

let rec reverse_fold_right f e = function
  | Nil -> e
  | Cons (head, tail) -> reverse_fold_right f (f head e) (force tail)
      
let rec reverse_fold_left f e = function
  | Nil -> e
  | Cons (head, tail) -> f (reverse_fold_left f e (force tail)) head


let cons x = function stream -> Cons (x, ref (Lazy.Value stream))

let rec snoc = fun stream x ->
  match stream with
    | Nil -> Cons (x, ref (Lazy.Value Nil))
    | Cons (head, tail) -> Cons (head, lazy (snoc (force tail) x))

let init = fun n f ->
  let rec init_rec = function
    | k when (k = n) -> Nil
    | k -> Cons (f k, lazy (init_rec (k + 1)))
  in
    if (n < 0) then raise (Invalid_argument "init : argument ngatif")
    else init_rec 0

let left_view = function
  | Nil -> raise Empty
  | Cons (head, tail) -> (head, force tail)

let left_head = function
  | Nil -> raise Empty
  | Cons (head, _) -> head

let left_tail = function
  | Nil -> raise Empty
  | Cons (_, tail) -> force tail

let right_view = function stream ->
  match stream with
  | Nil -> raise Empty
  | Cons (x, _) ->
      let right_head = ref x in
      let rec right_view_rec = function
	| Nil -> raise (Unexpected_pattern "Stream.right_view")
	| Cons (x, tail) -> 
	    match force tail with
	      | Nil -> right_head := x ; Nil
	      | other -> Cons (x, lazy (right_view_rec other))
      in
      let right_tail = right_view_rec stream in
	(right_tail, !right_head)

let rec right_tail = function
  | Nil -> raise Empty
  | Cons (head, tail) ->
      match force tail with
	| Nil -> Nil
	| stream -> Cons (head, lazy (right_tail stream))

let rec right_head = function
  | Nil -> raise Empty
  | Cons (head, tail) ->
      match force tail with
	| Nil -> head
	| stream -> right_head stream

let is_empty = function
  | Nil -> true
  | _ -> false

let rec size = function
  | Nil -> 0
  | Cons (_, tail) -> (1 + size (force tail))

let concat_map = fun f stream -> join (map f stream)

let rec fold_right f e = function
  | Nil -> e
  | Cons (head, tail) -> f head (fold_right f e (force tail))

let rec fold_left f e = function
  | Nil -> e
  | Cons (head, tail) -> fold_left f (f e head) (force tail)

let rec iter f = function
  | Nil -> ()
  | Cons (head, tail) -> f head ; iter f (force tail)

let rec reverse_iter f = function
  | Nil -> ()
  | Cons (head, tail) -> iter f (force tail) ; f head

let take n = function stream ->
  let rec take_rec k = function
    | _ when k = 0 -> Nil
    | Nil -> raise (Out_of_bounds n)
    | Cons (head, tail) -> Cons (head, lazy (take_rec (k - 1) (force tail)))
  in
    if n < 0 then raise (Invalid_argument "take : argument ngatif")
    else take_rec n stream

let drop n = function stream ->
  let rec drop_rec k = function
    | stream when k = 0 -> stream
    | Nil -> raise (Out_of_bounds n)
    | Cons (_, tail) -> drop_rec (k - 1) (force tail)
  in
    if n < 0 then raise (Invalid_argument "take : argument ngatif")
    else drop_rec n stream

let split_at n = function stream ->
  let rec split_rec k left = function
    | stream when k = 0 -> (from_list (List.reverse left), stream)
    | Cons (head, tail) -> split_rec (k - 1) (head :: left) (force tail)
    | _ -> raise (Out_of_bounds n)
  in
    split_rec n [] stream

let sub = fun starti endi stream ->
  let rec sub_rec n = function
    | Nil when n = endi -> Nil
    | Nil -> raise (Out_of_bounds endi)
    | Cons (_, tail) when n < starti -> sub_rec (n + 1) (force tail)
    | _ when n > endi -> Nil
    | Cons (head, tail) -> Cons (head, lazy (sub_rec (n + 1) (force tail)))
  in
    if (starti < 0) then raise (Invalid_argument "sub : argument ngatif")
    else if (starti > endi) then raise (Invalid_argument "sub : start > end")
    else sub_rec 0 stream

let rec exists f = function
  | Nil -> false
  | Cons (head, tail) -> f head || exists f (force tail)

let rec for_all f = function
  | Nil -> true
  | Cons (head, tail) -> f head && for_all f (force tail)
      
let rec find f = function
  | Nil -> raise Not_found
  | Cons (head, _) when f head -> head
  | Cons (_, tail) -> find f (force tail)

let rec filter f = function
  | Nil -> Nil
  | Cons (head, tail) when f head -> Cons (head, lazy (filter f (force tail)))
  | Cons (_, tail) -> filter f (force tail)

let rec partition f = function stream ->
  (filter f stream, filter (fun x -> not (f x)) stream)

let rec take_while f = function
  | Cons (head, tail) when f head -> Cons 
      (head, lazy (take_while f (force tail)))
  | _ -> Nil

let rec drop_while f = function
  | Cons (head, tail) when f head -> 
      Cons (head, lazy (drop_while f (force tail)))
  | stream -> stream

let rec split_rec f left = function
  | Cons (head, tail) when f head -> split_rec f (head :: left) (force tail)
  | right -> (from_list (List.reverse left), right)

let split_while = fun f stream -> split_rec f [] stream



(** Index based operations **)


let rec mapi_rec f k = function
  | Nil -> Nil
  | Cons (head, tail) -> 
      Cons (f k head, lazy (mapi_rec f (k + 1) (force tail)))

let map_with_index f = function stream -> mapi_rec f 0 stream

let rec iteri_rec f k = function
  | Nil -> ()
  | Cons (head, tail) -> f k head ; iteri_rec f (k + 1) (force tail)

let rec iter_with_index f = function stream -> iteri_rec f 0 stream

let rec findi_rec f k = function
  | Nil -> raise Not_found
  | Cons (head, _) when f k head -> head
  | Cons (_, tail) -> findi_rec f (k + 1) (force tail)

let find_with_index f = function stream -> findi_rec f 0 stream

let rec filteri_rec f k = function
  | Nil -> Nil
  | Cons (head, tail) when f k head -> 
      Cons (head, lazy (filteri_rec f (k + 1) (force tail)))
  | Cons (_, tail) -> filteri_rec f (k + 1) (force tail)

let filter_with_index f = function stream -> filteri_rec f 0 stream

let rec foldri f e k = function
  | Nil -> e
  | Cons (head, tail) -> f k head (foldri f e (k + 1) (force tail))

let fold_right_with_index f e = function stream -> foldri f e 0 stream

let rec foldli f e k = function
  | Nil -> e
  | Cons (head, tail) -> foldli f (f k e head) (k + 1) (force tail)

let fold_left_with_index f e = function stream -> foldli f e 0 stream

let rec zip = fun a b ->
  match (a, b) with
    | (Nil, Nil) -> Nil
    | (Cons (hda, tla) , Cons (hdb, tlb)) -> 
	Cons ((hda, hdb), lazy (zip (force tla) (force tlb)))
    | _ -> raise 
	(Invalid_argument "zip : les squences doivent tre de mme taille")

let rec unzip_left = function
  | Nil -> Nil
  | Cons ((a, _), tail) -> Cons (a, lazy (unzip_left (force tail)))

let rec unzip_right = function
  | Nil -> Nil
  | Cons ((_, b), tail) -> Cons (b, lazy (unzip_right (force tail)))

let unzip = function stream -> (unzip_left stream, unzip_right stream)


(** Aliases **)


let append = plus
let empty = zero

let blit = fun a b c d e -> raise (Not_working "blit")
let fill = fun a b c d -> raise (Not_working "fill")

module StreamSeq = 
struct

  type 'a seq = 'a stream

  let force = function lazyValue ->
    (match !lazyValue with
       | Lazy.Delayed _ -> print_endline "force" 
       | _ -> ()
    ) ;  
    Lazy.force lazyValue

  let rec count_forced = function
    | Nil -> 0
    | Cons (_, tail) ->
	match !tail with
	  | Lazy.Value _ -> 1 + count_forced (Lazy.force tail)
	  | _-> 1

  let unit = function x -> Cons (x, lazy Nil)
    
  let rec map f = function
    | Nil -> Nil
    | Cons (head, tail) -> Cons (f head, lazy (map f (force tail)))

  let rec plus = fun a b ->
    match a with
      | Nil -> b
      | Cons (head, tail) -> Cons (head, lazy (plus (force tail) b))

  let zero () = Nil

  let rec join = function
    | Nil -> Nil
    | Cons (head, tail) -> plus head (join (force tail))

  let bind = fun stream f -> join (map f stream)

  let rec make = fun n x ->
    match n with
      | 0 -> Nil
      | _ -> Cons (x, lazy (make (n - 1) x))

  let rec get = fun stream n ->
    match (stream, n) with
      | (Cons (x, _), 0) -> x
      | (Cons (_, tail), _) -> get (force tail) (n - 1)
      | _ -> raise Empty

  let rec set = fun stream n x ->
    match stream with
      | Nil when n = 0 -> Cons (x, lazy Nil)
      | Cons (_, tail) when n = 0 -> Cons (x, tail)
      | Cons (y, tail) -> Cons (y, lazy (set (force tail) (n - 1) x))
      | _ -> raise Empty

  (* From_list : les clotures sont directement forces 
   * puisque tous les lments de la liste sont connus 
   *)

  let rec from_list = function
    | [] -> Nil
    | head :: tail -> Cons (head, ref (Lazy.Value (from_list tail)))

  let rec to_list = function
    | Nil -> []
    | Cons (x, tail) -> x :: to_list (force tail)

  let rec to_list_append a = function
    | Nil -> a
    | Cons (x, tail) -> x :: to_list_append a (force tail)

  let rec reverse_onto result = function
    | Nil -> result
    | Cons (head, tail) -> 
	reverse_onto (Cons (head, ref (Lazy.Value result))) (force tail)

  let rec reverse_append = fun stream result ->
    match stream with
      | Nil -> result
      | Cons (head, tail) -> 
	  reverse_append (Cons (head, ref (Lazy.Value result))) (force tail)

  let reverse = function stream -> reverse_onto Nil stream

  let rec append_reverse_rec result = function
    | (Nil, stream) -> reverse_onto result stream
    | (stream, Nil) -> stream ++ result
    | (Cons (x, xs), Cons (y, ys)) ->
	Cons (x, lazy (append_reverse_rec (Cons (y, lazy result)) 
			 (force xs, force ys)))

  let rec append_reverse = fun a b -> append_reverse_rec Nil (a, b)

  let rec reverse_iter f = function
    | Nil -> ()
    | Cons (head, tail) -> reverse_iter f (force tail) ; f head

  let rec reverse_fold_right f e = function
    | Nil -> e
    | Cons (head, tail) -> reverse_fold_right f (f head e) (force tail)
	
  let rec reverse_fold_left f e = function
    | Nil -> e
    | Cons (head, tail) -> f (reverse_fold_left f e (force tail)) head


  let cons x = function stream -> Cons (x, ref (Lazy.Value stream))

  let rec snoc = fun stream x ->
    match stream with
      | Nil -> Cons (x, ref (Lazy.Value Nil))
      | Cons (head, tail) -> Cons (head, lazy (snoc (force tail) x))

  let init = fun n f ->
    let rec init_rec = function
      | k when (k = n) -> Nil
      | k -> Cons (f k, lazy (init_rec (k + 1)))
    in
      if (n < 0) then raise (Invalid_argument "init : argument ngatif")
      else init_rec 0

  let left_view = function
    | Nil -> raise Empty
    | Cons (head, tail) -> (head, force tail)

  let left_head = function
    | Nil -> raise Empty
    | Cons (head, _) -> head

  let left_tail = function
    | Nil -> raise Empty
    | Cons (_, tail) -> force tail

  let right_view = function stream ->
    match stream with
      | Nil -> raise Empty
      | Cons (x, _) ->
	  let right_head = ref x in
	  let rec right_view_rec = function
	    | Nil -> raise (Unexpected_pattern "Stream.right_view")
	    | Cons (x, tail) -> 
		match force tail with
		  | Nil -> right_head := x ; Nil
		  | other -> Cons (x, lazy (right_view_rec other))
	  in
	  let right_tail = right_view_rec stream in
	    (right_tail, !right_head)

  let rec right_tail = function
    | Nil -> raise Empty
    | Cons (head, tail) ->
	match force tail with
	  | Nil -> Nil
	  | stream -> Cons (head, lazy (right_tail stream))

  let rec right_head = function
    | Nil -> raise Empty
    | Cons (head, tail) ->
	match force tail with
	  | Nil -> head
	  | stream -> right_head stream

  let is_empty = function
    | Nil -> true
    | _ -> false

  let rec size = function
    | Nil -> 0
    | Cons (_, tail) -> (1 + size (force tail))

  let concat_map = fun f stream -> join (map f stream)

  let rec fold_right f e = function
    | Nil -> e
    | Cons (head, tail) -> f head (fold_right f e (force tail))

  let rec fold_left f e = function
    | Nil -> e
    | Cons (head, tail) -> fold_left f (f e head) (force tail)

  let rec iter f = function
    | Nil -> ()
    | Cons (head, tail) -> f head ; iter f (force tail)

  let rec reverse_iter f = function
    | Nil -> ()
    | Cons (head, tail) -> iter f (force tail) ; f head

  let take n = function stream ->
    let rec take_rec k = function
      | _ when k = 0 -> Nil
      | Nil -> raise (Out_of_bounds n)
      | Cons (head, tail) -> Cons (head, lazy (take_rec (k - 1) (force tail)))
    in
      if n < 0 then raise (Invalid_argument "take : argument ngatif")
      else take_rec n stream

  let drop n = function stream ->
    let rec drop_rec k = function
      | stream when k = 0 -> stream
      | Nil -> raise (Out_of_bounds n)
      | Cons (_, tail) -> drop_rec (k - 1) (force tail)
    in
      if n < 0 then raise (Invalid_argument "take : argument ngatif")
      else drop_rec n stream

  let split_at n = function stream ->
    let rec split_rec k left = function
      | stream when k = 0 -> (from_list (List.reverse left), stream)
      | Cons (head, tail) -> split_rec (k - 1) (head :: left) (force tail)
      | _ -> raise (Out_of_bounds n)
    in
      split_rec n [] stream

  let sub = fun starti endi stream ->
    let rec sub_rec n = function
      | Nil when n = endi -> Nil
      | Nil -> raise (Out_of_bounds endi)
      | Cons (_, tail) when n < starti -> sub_rec (n + 1) (force tail)
      | _ when n > endi -> Nil
      | Cons (head, tail) -> Cons (head, lazy (sub_rec (n + 1) (force tail)))
    in
      if (starti < 0) then raise (Invalid_argument "sub : argument ngatif")
      else if (starti > endi) then raise (Invalid_argument "sub : start > end")
      else sub_rec 0 stream

  let rec exists f = function
    | Nil -> false
    | Cons (head, tail) -> f head || exists f (force tail)

  let rec for_all f = function
    | Nil -> true
    | Cons (head, tail) -> f head && for_all f (force tail)
	
  let rec find f = function
    | Nil -> raise Not_found
    | Cons (head, _) when f head -> head
    | Cons (_, tail) -> find f (force tail)

  let rec filter f = function
    | Nil -> Nil
    | Cons (head, tail) when f head ->
	Cons (head, lazy (filter f (force tail)))
    | Cons (_, tail) -> filter f (force tail)

  let rec partition f = function stream ->
    (filter f stream, filter (fun x -> not (f x)) stream)

  let rec take_while f = function
    | Cons (head, tail) when f head -> Cons 
	(head, lazy (take_while f (force tail)))
    | _ -> Nil

  let rec drop_while f = function
    | Cons (head, tail) when f head -> 
	Cons (head, lazy (drop_while f (force tail)))
    | stream -> stream

  let rec split_rec f left = function
    | Cons (head, tail) when f head -> split_rec f (head :: left) (force tail)
    | right -> (from_list (List.reverse left), right)

  let split_while = fun f stream -> split_rec f [] stream



  (** Index based operations **)


  let rec mapi_rec f k = function
    | Nil -> Nil
    | Cons (head, tail) -> 
	Cons (f k head, lazy (mapi_rec f (k + 1) (force tail)))

  let map_with_index f = function stream -> mapi_rec f 0 stream

  let rec iteri_rec f k = function
    | Nil -> ()
    | Cons (head, tail) -> f k head ; iteri_rec f (k + 1) (force tail)

  let rec iter_with_index f = function stream -> iteri_rec f 0 stream

  let rec findi_rec f k = function
    | Nil -> raise Not_found
    | Cons (head, _) when f k head -> head
    | Cons (_, tail) -> findi_rec f (k + 1) (force tail)

  let find_with_index f = function stream -> findi_rec f 0 stream

  let rec filteri_rec f k = function
    | Nil -> Nil
    | Cons (head, tail) when f k head -> 
	Cons (head, lazy (filteri_rec f (k + 1) (force tail)))
    | Cons (_, tail) -> filteri_rec f (k + 1) (force tail)

  let filter_with_index f = function stream -> filteri_rec f 0 stream

  let rec foldri f e k = function
    | Nil -> e
    | Cons (head, tail) -> f k head (foldri f e (k + 1) (force tail))

  let fold_right_with_index f e = function stream -> foldri f e 0 stream

  let rec foldli f e k = function
    | Nil -> e
    | Cons (head, tail) -> foldli f (f k e head) (k + 1) (force tail)

  let fold_left_with_index f e = function stream -> foldli f e 0 stream

  let rec zip = fun a b ->
    match (a, b) with
      | (Nil, Nil) -> Nil
      | (Cons (hda, tla) , Cons (hdb, tlb)) -> 
	  Cons ((hda, hdb), lazy (zip (force tla) (force tlb)))
      | _ -> raise 
	  (Invalid_argument "zip : les squences doivent tre de mme taille")

  let rec unzip_left = function
    | Nil -> Nil
    | Cons ((a, _), tail) -> Cons (a, lazy (unzip_left (force tail)))

  let rec unzip_right = function
    | Nil -> Nil
    | Cons ((_, b), tail) -> Cons (b, lazy (unzip_right (force tail)))

  let unzip = function stream -> (unzip_left stream, unzip_right stream)


  (** Aliases **)


  let append = plus
  let empty = zero

  let blit = fun a b c d e -> raise (Not_working "blit")
  let fill = fun a b c d -> raise (Not_working "fill")

end
