open EdisonPervasives
 
type 
  'a lazyCell = Nil | Cons of 'a * 'a stream and
  'a stream = 'a lazyCell Lazy.t

let unit = function x -> lazy (Cons (x, lazy Nil))
    
let rec map f = function stream ->
  match force stream with
    | Nil -> lazy Nil
    | Cons (head, tail) -> lazy (Cons (f head, map f tail))

let rec plus = fun a b ->
  match force a with
    | Nil -> b
    | Cons (head, tail) -> lazy (Cons (head, plus tail b))

let zero () = lazy Nil

let rec join = function stream ->
  match force stream with
    | Nil -> lazy Nil
    | Cons (head, tail) -> plus head (join tail)

let bind = fun stream f -> join (map f stream)

let rec make = fun n x ->
  match n with
    | 0 -> lazy Nil
    | _ -> lazy (Cons (x, make (n - 1) x))

let rec get = fun stream n ->
  match force stream with
    | Cons (x, _) when n = 0 -> x
    | Cons (_, tail)  -> get tail (n - 1)
    | _ -> raise Empty

let rec set = fun stream n x ->
  match force stream with
    | Nil when n = 0 -> lazy (Cons (x, lazy Nil))
    | Cons (_, tail) when n = 0 -> lazy (Cons (x, tail))
    | Cons (y, tail) -> lazy (Cons (y, set tail (n - 1) x))
    | _ -> raise Empty

(* From_list : les clotures sont directement forces 
 * puisque tous les lments de la liste sont connus 
 *)

let rec from_list_rec = function
  | [] -> Nil
  | head :: tail -> Cons (head, strict (from_list_rec tail))

let from_list = function list -> lazy (from_list_rec list)

let rec to_list = function stream ->
  match force stream with
    | Nil -> []
    | Cons (x, tail) -> x :: to_list tail

let rec to_list_append a = function stream ->
  match force stream with
    | Nil -> a
    | Cons (x, tail) -> x :: to_list_append a tail

let rec reverse_onto result = function stream ->
  match force stream with
    | Nil -> result
    | Cons (head, tail) -> 
	reverse_onto (strict (Cons (head, result))) tail

let rec reverse_append = fun stream result ->
  match force stream with
    | Nil -> result
    | Cons (head, tail) -> 
	reverse_append (strict (Cons (head, result))) tail

let reverse = function stream -> reverse_onto (lazy Nil) stream

let rec append_reverse_rec result = fun a b ->
  match (force a, force b) with
    | (Nil, stream) -> reverse_onto result (lazy stream)
    | (stream, Nil) -> plus (lazy stream) result
    | (Cons (x, xs), Cons (y, ys)) ->
	lazy (Cons (x, append_reverse_rec (lazy (Cons (y, result))) xs ys))

let rec append_reverse = fun a b -> append_reverse_rec (lazy Nil) a b

let rec reverse_iter f = function stream ->
  match force stream with
    | Nil -> ()
    | Cons (head, tail) -> reverse_iter f tail ; f head

let cons x = function stream -> lazy (Cons (x, stream))

let rec snoc = fun stream x ->
  match force stream with
    | Nil -> lazy (Cons (x, lazy Nil))
    | Cons (head, tail) -> lazy (Cons (head, snoc tail x))

let init = fun n f ->
  let rec init_rec = function
    | k when (k = n) -> lazy Nil
    | k -> lazy (Cons (f k, init_rec (k + 1)))
  in
    init_rec 0

let left_view = function stream ->
  match force stream with
    | Nil -> raise Empty
    | Cons (head, tail) -> (head, tail)

let left_head = function stream ->
  match force stream with
    | Nil -> raise Empty
    | Cons (head, _) -> head

let left_tail = function stream ->
  match force stream with
    | Nil -> raise Empty
    | Cons (_, tail) -> tail

let right_view = function stream ->
  match force stream with
    | Nil -> raise Empty
    | Cons (x, _) ->
	let right_head = ref x in
	let rec right_view_rec = function
	  | Nil -> raise (Unexpected_pattern "Stream.right_view")
	  | Cons (x, tail) -> 
	      match force tail with
		| Nil -> right_head := x ; strict Nil
		| other -> strict (Cons (x, right_view_rec other))
	in
	let right_tail = right_view_rec (force stream) in
	  (right_tail, !right_head)

let rec right_tail_rec = function
  | Nil -> raise Empty
  | Cons (head, tail) ->
      match force tail with
	| Nil -> strict Nil
	| streamCell -> strict (Cons (head, right_tail_rec streamCell))

let right_tail = function stream -> right_tail_rec (force stream)

let rec right_head_rec = function
  | Nil -> raise Empty
  | Cons (head, tail) ->
      match force tail with
	| Nil -> head
	| stream -> right_head_rec stream

let right_head = function stream -> right_head_rec (force stream)
	    
let is_empty = function stream ->
  match force stream with
    | Nil -> true
    | _ -> false

let rec size = function stream ->
  match force stream with
    | Nil -> 0
    | Cons (_, tail) -> 1 + size tail

let concat_map = fun f stream -> join (map f stream)

let rec fold_right f e = function stream ->
  match force stream with
    | Nil -> e
    | Cons (head, tail) -> f head (fold_right f e tail)

let rec fold_left f e = function stream ->
  match force stream with
    | Nil -> e
    | Cons (head, tail) -> fold_left f (f e head) tail
	
let rec iter f = function stream ->
  match force stream with
    | Nil -> ()
    | Cons (head, tail) -> f head ; iter f tail

let rec reverse_iter f = function stream ->
  match force stream with
    | Nil -> ()
    | Cons (head, tail) -> reverse_iter f tail ; f head
	
let take n = function str ->
  let rec take_rec k = function stream ->
    if k = 0 then lazy Nil
    else match force stream with
      | Nil -> raise (Out_of_bounds n)
      | Cons (head, tail) -> lazy (Cons (head, take_rec (k - 1) tail))
  in
    if n < 0 then raise (Invalid_argument "take : argument ngatif")
    else take_rec n str

let drop n = function stream ->
  let rec drop_rec k = function stream ->
    match force stream with
      | _ when k = 0 -> stream
      | Nil -> raise (Out_of_bounds n)
      | Cons (_, tail) -> drop_rec (k - 1) tail
  in
    if n < 0 then raise (Invalid_argument "take : argument ngatif")
    else drop_rec n stream

let lazy_drop n = function stream ->
  let rec drop_rec k = function stream ->
    match force stream with
      | streamCell when k = 0 -> streamCell
      | Nil -> raise (Out_of_bounds k)
      | Cons (_, tail) -> drop_rec (k - 1) tail
  in
    if n < 0 then raise (Invalid_argument "take : argument ngatif")
    else lazy (drop_rec n stream)


let split_at n = function stream ->
  let rec split_rec k left = function stream ->
    match force stream with
      | _ when k = 0 -> (from_list (List.reverse left), stream)
      | Cons (head, tail) -> split_rec (k - 1) (head :: left) tail
      | _ -> raise (Out_of_bounds n)
  in
    split_rec n [] stream

let sub = fun starti endi stream ->
  let rec sub_rec n = function stream ->
    match force stream with
      | Nil when n = endi -> Nil
      | Nil -> raise (Out_of_bounds endi)
      | Cons (_, tail) when n < starti -> sub_rec (n + 1) tail
      | _ when n > endi -> Nil
      | Cons (head, tail) -> Cons (head, lazy (sub_rec (n + 1) tail))
  in
    if (starti < 0) then raise (Invalid_argument "sub : argument ngatif")
    else if (starti > endi) then raise (Invalid_argument "sub : start > end")
    else lazy (sub_rec 0 stream)
      
let rec exists f = function stream ->
  match force stream with
    | Nil -> false
    | Cons (head, tail) -> f head || exists f tail

let rec for_all f = function stream ->
  match force stream with
    | Nil -> true
    | Cons (head, tail) -> f head && for_all f tail
      
let rec find f = function stream ->
  match force stream with
    | Nil -> raise Not_found
    | Cons (head, _) when f head -> head
    | Cons (_, tail) -> find f tail

let rec filter f = function stream ->
  match force stream with
    | Nil -> lazy Nil
    | Cons (head, tail) when f head -> lazy (Cons (head, filter f tail))
    | Cons (_, tail) -> filter f tail

let rec partition f = function stream ->
  (filter f stream, filter (fun x -> not (f x)) stream)

let rec take_while f = function stream ->
  match force stream with
    | Cons (head, tail) when f head -> lazy (Cons (head, take_while f tail))
    | _ -> lazy Nil

let rec drop_while f = function stream ->
  match force stream with
    | Cons (head, tail) when f head -> lazy (Cons (head, drop_while f tail))
    | _ -> stream

let rec split_rec f left = function stream ->
  match force stream with
    | Cons (head, tail) when f head -> split_rec f (head :: left) tail
    | _ -> (from_list (List.reverse left), stream)

let split_while = fun f stream -> split_rec f [] stream



(** Index based operations **)


let rec mapi_rec f k = function stream ->
  match force stream with
    | Nil -> lazy Nil
    | Cons (head, tail) -> lazy (Cons (f k head, mapi_rec f (k + 1) tail))

let map_with_index f = function stream -> mapi_rec f 0 stream

let rec iteri_rec f k = function stream ->
  match force stream with
    | Nil -> ()
    | Cons (head, tail) -> f k head ; iteri_rec f (k + 1) tail

let rec iter_with_index f = function stream -> iteri_rec f 0 stream

let rec findi_rec f k = function stream ->
  match force stream with
    | Nil -> raise Not_found
    | Cons (head, _) when f k head -> head
    | Cons (_, tail) -> findi_rec f (k + 1) tail

let find_with_index f = function stream -> findi_rec f 0 stream

let rec filteri_rec f k = function stream ->
  match force stream with
    | Nil -> lazy Nil
    | Cons (head, tail) when f k head -> 
	lazy (Cons (head, filteri_rec f (k + 1) tail))
    | Cons (_, tail) -> filteri_rec f (k + 1) tail

let filter_with_index f = function stream -> filteri_rec f 0 stream

let rec foldri f e k = function stream ->
  match force stream with
    | Nil -> e
    | Cons (head, tail) -> f k head (foldri f e (k + 1) tail)

let fold_right_with_index f e = function stream -> foldri f e 0 stream

let rec foldli f e k = function stream ->
  match force stream with
    | Nil -> e
    | Cons (head, tail) -> foldli f (f k e head) (k + 1) tail

let fold_left_with_index f e = function stream -> foldli f e 0 stream

let rec zip = fun a b ->
  match (force a, force b) with
    | (Nil, Nil) -> lazy Nil
    | (Cons (hda, tla) , Cons (hdb, tlb)) -> 
	lazy (Cons ((hda, hdb), zip tla tlb))
    | _ -> raise 
	(Invalid_argument "zip : les squences doivent tre de mme taille")

let rec unzip_left = function stream ->
  match force stream with
    | Nil -> lazy Nil
    | Cons ((a, _), tail) -> lazy (Cons (a, unzip_left tail))

let rec unzip_right = function stream ->
  match force stream with
    | Nil -> lazy Nil
    | Cons ((_, b), tail) -> lazy (Cons (b, unzip_right tail))

let unzip = function stream -> (unzip_left stream, unzip_right stream)

let empty = zero
let append = plus
	       
let blit = fun a b c d e -> raise (Not_working "blit")
let fill = fun a b c d -> raise (Not_working "fill")
  


module StreamSeq = struct

  type 
    'a lazyCell = Nil | Cons of 'a * 'a seq and
    'a seq = 'a lazyCell Lazy.t

  let strict = function x -> ref (Lazy.Value x)

  let unit = function x -> lazy (Cons (x, lazy Nil))
    
  let rec map f = function stream ->
    match force stream with
      | Nil -> lazy Nil
      | Cons (head, tail) -> lazy (Cons (f head, map f tail))

  let rec plus = fun a b ->
    match force a with
      | Nil -> b
      | Cons (head, tail) -> lazy (Cons (head, plus tail b))

  let zero () = lazy Nil

  let rec join = function stream ->
    match force stream with
      | Nil -> lazy Nil
      | Cons (head, tail) -> plus head (join tail)

  let bind = fun stream f -> join (map f stream)

  let rec make = fun n x ->
    match n with
      | 0 -> lazy Nil
      | _ -> lazy (Cons (x, make (n - 1) x))

  let rec get = fun stream n ->
    match (force stream, n) with
      | (Cons (x, _), 0) -> x
      | (Cons (_, tail), _) -> get tail (n - 1)
      | _ -> raise Empty

  let rec set = fun stream n x ->
    match force stream with
      | Nil when n = 0 -> lazy (Cons (x, lazy Nil))
      | Cons (_, tail) when n = 0 -> lazy (Cons (x, tail))
      | Cons (y, tail) -> lazy (Cons (y, set tail (n - 1) x))
      | _ -> raise Empty

  (* From_list : les clotures sont directement forces 
   * puisque tous les lments de la liste sont connus 
   *)

  let rec from_list_rec = function
    | [] -> Nil
    | head :: tail -> Cons (head, strict (from_list_rec tail))

  let from_list = function list -> lazy (from_list_rec list)

  let rec to_list = function stream ->
    match force stream with
      | Nil -> []
      | Cons (x, tail) -> x :: to_list tail

  let rec to_list_append a = function stream ->
    match force stream with
      | Nil -> a
      | Cons (x, tail) -> x :: to_list_append a tail

  let rec reverse_onto result = function stream ->
    match force stream with
      | Nil -> result
      | Cons (head, tail) -> 
	  reverse_onto (strict (Cons (head, result))) tail

  let rec reverse_append = fun stream result ->
    match force stream with
      | Nil -> result
      | Cons (head, tail) -> 
	  reverse_append (strict (Cons (head, result))) tail

  let reverse = function stream -> reverse_onto (lazy Nil) stream

  let rec append_reverse_rec result = fun a b ->
    match (force a, force b) with
      | (Nil, stream) -> reverse_onto result (lazy stream)
      | (stream, Nil) -> plus (lazy stream) result
      | (Cons (x, xs), Cons (y, ys)) ->
	  lazy (Cons (x, append_reverse_rec (lazy (Cons (y, result))) xs ys))

  let rec append_reverse = fun a b -> append_reverse_rec (lazy Nil) a b

  let rec reverse_iter f = function stream ->
    match force stream with
      | Nil -> ()
      | Cons (head, tail) -> reverse_iter f tail ; f head

  let cons x = function stream -> lazy (Cons (x, stream))

  let rec snoc = fun stream x ->
    match force stream with
      | Nil -> lazy (Cons (x, lazy Nil))
      | Cons (head, tail) -> lazy (Cons (head, snoc tail x))

  let init = fun n f ->
    let rec init_rec = function
      | k when (k = n) -> lazy Nil
      | k -> lazy (Cons (f k, init_rec (k + 1)))
    in
      if (n < 0) then raise (Invalid_argument "init : argument ngatif")
      else init_rec 0

  let left_view = function stream ->
    match force stream with
      | Nil -> raise Empty
      | Cons (head, tail) -> (head, tail)

  let left_head = function stream ->
    match force stream with
      | Nil -> raise Empty
      | Cons (head, _) -> head

  let left_tail = function stream ->
    match force stream with
      | Nil -> raise Empty
      | Cons (_, tail) -> tail

  let right_view = function stream ->
    match force stream with
      | Nil -> raise Empty
      | Cons (x, _) ->
	  let right_head = ref x in
	  let rec right_view_rec = function
	    | Nil -> raise (Unexpected_pattern "Stream.right_view")
	    | Cons (x, tail) -> 
		match force tail with
		  | Nil -> right_head := x ; strict Nil
		  | other -> strict (Cons (x, right_view_rec other))
	  in
	  let right_tail = right_view_rec (force stream) in
	    (right_tail, !right_head)

  let rec right_tail_rec = function
    | Nil -> raise Empty
    | Cons (head, tail) ->
	match force tail with
	  | Nil -> strict Nil
	  | streamCell -> strict (Cons (head, right_tail_rec streamCell))

  let right_tail = function stream -> right_tail_rec (force stream)

  let rec right_head_rec = function
    | Nil -> raise Empty
    | Cons (head, tail) ->
	match force tail with
	  | Nil -> head
	  | stream -> right_head_rec stream

  let right_head = function stream -> right_head_rec (force stream)
    
  let is_empty = function stream ->
    match force stream with
      | Nil -> true
      | _ -> false

  let rec size = function stream ->
    match force stream with
      | Nil -> 0
      | Cons (_, tail) -> (1 + size tail)

  let concat_map = fun f stream -> join (map f stream)

  let rec fold_right f e = function stream ->
    match force stream with
      | Nil -> e
      | Cons (head, tail) -> f head (fold_right f e tail)

  let rec fold_left f e = function stream ->
    match force stream with
      | Nil -> e
      | Cons (head, tail) -> fold_left f (f e head) tail
	  
  let rec iter f = function stream ->
    match force stream with
      | Nil -> ()
      | Cons (head, tail) -> f head ; iter f tail

  let rec reverse_iter f = function stream ->
    match force stream with
      | Nil -> ()
      | Cons (head, tail) -> iter f tail ; f head
	  
  let take n = function str ->
    let rec take_rec k = function stream ->
      if k = 0 then lazy Nil
      else match force stream with
	| Nil -> raise (Out_of_bounds n)
	| Cons (head, tail) -> lazy (Cons (head, take_rec (k - 1) tail))
    in
      if n < 0 then raise (Invalid_argument "take : argument ngatif")
      else take_rec n str

  let drop n = function stream ->
    let rec drop_rec k = function stream ->
      match force stream with
	| _ when k = 0 -> stream
	| Nil -> raise (Out_of_bounds n)
	| Cons (_, tail) -> drop_rec (k - 1) tail
    in
      if n < 0 then raise (Invalid_argument "take : argument ngatif")
      else drop_rec n stream

  let lazy_drop n = function stream ->
    let rec drop_rec k = function stream ->
      match force stream with
	| streamCell when k = 0 -> streamCell
	| Nil -> raise (Out_of_bounds k)
	| Cons (_, tail) -> drop_rec (k - 1) tail
    in
      if n < 0 then raise (Invalid_argument "take : argument ngatif")
      else lazy (drop_rec n stream)


  let split_at n = function stream ->
    let rec split_rec k left = function stream ->
      match force stream with
	| _ when k = 0 -> (from_list (List.reverse left), stream)
	| Cons (head, tail) -> split_rec (k - 1) (head :: left) tail
	| _ -> raise (Out_of_bounds n)
    in
      split_rec n [] stream

  let sub = fun starti endi stream ->
    let rec sub_rec n = function stream ->
      match force stream with
	| Nil when n = endi -> Nil
	| Nil -> raise (Out_of_bounds endi)
	| Cons (_, tail) when n < starti -> sub_rec (n + 1) tail
	| _ when n > endi -> Nil
	| Cons (head, tail) -> Cons (head, lazy (sub_rec (n + 1) tail))
    in
      if (starti < 0) then raise (Invalid_argument "sub : argument ngatif")
      else if (starti > endi) then raise (Invalid_argument "sub : start > end")
      else lazy (sub_rec 0 stream)
	
  let rec exists f = function stream ->
    match force stream with
      | Nil -> false
      | Cons (head, tail) -> f head || exists f tail

  let rec for_all f = function stream ->
    match force stream with
      | Nil -> true
      | Cons (head, tail) -> f head && for_all f tail
	  
  let rec find f = function stream ->
    match force stream with
      | Nil -> raise Not_found
      | Cons (head, _) when f head -> head
      | Cons (_, tail) -> find f tail

  let rec filter f = function stream ->
    match force stream with
      | Nil -> lazy Nil
      | Cons (head, tail) when f head -> lazy (Cons (head, filter f tail))
      | Cons (_, tail) -> filter f tail

  let rec partition f = function stream ->
    (filter f stream, filter (fun x -> not (f x)) stream)

  let rec take_while f = function stream ->
    match force stream with
      | Cons (head, tail) when f head -> lazy (Cons (head, take_while f tail))
      | _ -> lazy Nil

  let rec drop_while f = function stream ->
    match force stream with
      | Cons (head, tail) when f head -> lazy (Cons (head, drop_while f tail))
      | _ -> stream

  let rec split_rec f left = function stream ->
    match force stream with
      | Cons (head, tail) when f head -> split_rec f (head :: left) tail
      | _ -> (from_list (List.reverse left), stream)

  let split_while = fun f stream -> split_rec f [] stream



  (** Index based operations **)


  let rec mapi_rec f k = function stream ->
    match force stream with
      | Nil -> lazy Nil
      | Cons (head, tail) -> lazy (Cons (f k head, mapi_rec f (k + 1) tail))

  let map_with_index f = function stream -> mapi_rec f 0 stream

  let rec iteri_rec f k = function stream ->
    match force stream with
      | Nil -> ()
      | Cons (head, tail) -> f k head ; iteri_rec f (k + 1) tail

  let rec iter_with_index f = function stream -> iteri_rec f 0 stream

  let rec findi_rec f k = function stream ->
    match force stream with
      | Nil -> raise Not_found
      | Cons (head, _) when f k head -> head
      | Cons (_, tail) -> findi_rec f (k + 1) tail

  let find_with_index f = function stream -> findi_rec f 0 stream

  let rec filteri_rec f k = function stream ->
    match force stream with
      | Nil -> lazy Nil
      | Cons (head, tail) when f k head -> 
	  lazy (Cons (head, filteri_rec f (k + 1) tail))
      | Cons (_, tail) -> filteri_rec f (k + 1) tail

  let filter_with_index f = function stream -> filteri_rec f 0 stream

  let rec foldri f e k = function stream ->
    match force stream with
      | Nil -> e
      | Cons (head, tail) -> f k head (foldri f e (k + 1) tail)

  let fold_right_with_index f e = function stream -> foldri f e 0 stream

  let rec foldli f e k = function stream ->
    match force stream with
      | Nil -> e
      | Cons (head, tail) -> foldli f (f k e head) (k + 1) tail

  let fold_left_with_index f e = function stream -> foldli f e 0 stream

  let rec zip = fun a b ->
    match (force a, force b) with
      | (Nil, Nil) -> lazy Nil
      | (Cons (hda, tla) , Cons (hdb, tlb)) -> 
	  lazy (Cons ((hda, hdb), zip tla tlb))
      | _ -> raise 
	  (Invalid_argument "zip : les squences doivent tre de mme taille")

  let rec unzip_left = function stream ->
    match force stream with
      | Nil -> lazy Nil
      | Cons ((a, _), tail) -> lazy (Cons (a, unzip_left tail))

  let rec unzip_right = function stream ->
    match force stream with
      | Nil -> lazy Nil
      | Cons ((_, b), tail) -> lazy (Cons (b, unzip_right tail))

  let unzip = function stream -> (unzip_left stream, unzip_right stream)

  let empty = zero
  let append = plus

  let blit = fun a b c d e -> raise (Not_working "blit")
  let fill = fun a b c d -> raise (Not_working "fill")

end
