(* Streams without memoization *)

(* Ceci est un module  fins essentiellement pdagogiques *)

open EdisonPervasives

type 'a stream =
  | Nil
  | Cons of 'a * 'a stream
  | Delayed of (unit -> 'a stream)

let lazy_ = function f -> Delayed (function () -> f)

let rec force = function
  | Delayed f -> print_endline "force" ; force (f ())
  | stream -> stream

(* la fonction [lazy_] ne fait qu'muler partiellement la construction  
 * [lazy] de Caml, la diffrence tant dans le fait qu'une fonction 
 * value toujours ses arguments avant de les utiliser.
 * Par contre, ni [lazy exp], ni [function () -> exp] ne vont valuer [exp]

   # let rec range = function n -> Cons (n, lazy_ (range (n + 1))) ;;
   val range : int -> 'a stream
   # let x = range 0 ;;
   Stack overflow during evaluation (looping recursion?).

 * en effet, [lazy_] a valu [range (n + 1)]
   
   # let rec range = function n -> 
   Cons (n, Delayed (fun () -> range (n + 1))) ;;
   val range : int -> int stream = <fun>
   # let x = range 0 ;;
   val x : int stream = Cons (0, Delayed <fun>)

   # let rec range = function n -> 
   lazy_ (Cons (n, Delayed (fun () -> range (n + 1)))) ;;
   val range : int -> int stream = <fun>
   # let x = range 0 ;;
   val x : int stream = Delayed <fun>

   # let rec range = function n -> 
   Delayed (fun () -> Cons (n, lazy_ (range (n + 1)))) ;;
   val range : int -> int stream = <fun>
   # let x = range 0;;
   val x : int stream = Delayed <fun>

 *)

(* La mthode de construction des fonctions suivantes est trs simple :
 *  - pour Nil et Cons (head, tail) les fonctions se comportent
 *  exactement comme les fonctions respectives du module List
 *  - on rajoute un troisime cas de la forme
 *  | stream -> Delay (function () -> [nom_de_fonction] (force stream))

 * Nous avons en outre rajout quelques fonctions suffixes par _lazy qui
 * construisent systmatiquement des suspensions 
 *)


let unit = function x -> Cons (x, Nil)

let rec map f = function
  | Nil -> Nil
  | Cons (x, stream) -> Cons (f x, map f stream)
  | stream -> Delayed (function () -> map f (force stream))

let rec plus = fun a b ->
  match a with
    | Nil -> b
    | Cons (x, stream) -> Cons (x, plus stream b)
    | stream -> Delayed (function () -> plus (force stream) b)

let rec join = function
  | Nil -> Nil
  | Cons (head, tail) -> plus head (join tail)
  | stream -> Delayed (function () -> join (force stream))

let rec bind = fun stream f ->
  match stream with
    | Nil -> Nil
    | Cons (head, tail) -> plus (f head) (bind tail f)
    | stream -> Delayed (function () -> bind (force stream) f)

let rec make = fun n x ->
  if n = 0 then Nil else Cons (x, make (n - 1) x)

let rec make_lazy = fun n x ->
  if n = 0 then Nil
  else Delayed (fun () -> Cons (x, make_lazy (n - 1) x))

let rec get = fun stream n ->
  match force stream with
    | Nil -> raise (Out_of_bounds n)
    | Cons (head, _) when n = 0 -> head
    | Cons (_, tail) -> get tail (n - 1)
    | _ -> raise (Unexpected_pattern "StreamIII.get")

let rec set = fun stream n x ->
  match stream with
    | Nil -> raise (Out_of_bounds n)
    | Cons (_, tail) when n = 0 -> Cons (x, tail)
    | Cons (head, tail) -> Cons (head, set tail (n - 1) x)
    | _ -> Delayed (function () -> set (force stream) n x)

let rec reverse_onto result = function
  | Nil -> result
  | Cons (head, tail) -> reverse_onto (Cons (head, result)) tail
  | stream -> Delayed (function () -> reverse_onto result (force stream))

let reverse = function stream -> reverse_onto Nil stream

let cons x = function stream -> Cons (x, stream)

let rec from_list = function
  | [] -> Nil
  | head :: tail -> Cons (head, from_list tail)

let rec from_list_lazy = function
  | [] -> lazy_ Nil
  | head :: tail -> lazy_ (Cons (head, from_list_lazy tail))

let init = fun n f ->
  let rec init_rec = function
    | k when (k = n) -> Nil
    | k -> Delayed (function () -> Cons (f k, init_rec (k + 1)))
  in
    init_rec 0

let left_view = function stream ->
  match force stream with
    | Nil -> raise Empty
    | Cons (head, tail) -> (head, tail)
    | _ -> raise (Unexpected_pattern "StreamIII.left_view")

let left_head = function stream ->
  match force stream with
    | Nil -> raise Empty
    | Cons (head, _) -> head
    | _ -> raise (Unexpected_pattern "StreamIII.left_head")

let left_tail = function stream ->
  match force stream with
    | Nil -> raise Empty
    | Cons (_, tail) -> tail
    | _ -> raise (Unexpected_pattern "StreamIII.left_tail")

let rec is_empty = function stream ->
  match force stream with
    | Nil -> true
    | _ -> false

let rec fold_right f e = function stream ->
  match force stream with
    | Nil -> e
    | Cons (head, tail) -> f head (fold_right f e tail)
    | _ -> raise (Unexpected_pattern "StreamIII.fold_right")

let rec fold_left f e = function stream ->
  match force stream with
    | Nil -> e
    | Cons (head, tail) -> fold_left f (f e head) tail
    | _ -> raise (Unexpected_pattern "StreamIII.fold_left")

let rec iter f = function stream ->
  match force stream with
    | Nil -> ()
    | Cons (head, tail) -> f head ; iter f tail
    | _ -> raise (Unexpected_pattern "StreamIII.iter")

let rec take n = function stream -> 
  if n = 0 then Nil
  else match stream with
    | Nil -> raise (Out_of_bounds n)
    | Cons (x, tail) -> Cons (x, take (n - 1) tail)
    | stream -> Delayed (function () -> take (n - 1) (force stream))

let rec drop n = function stream ->
  if n = 0 then stream
  else match stream with
    | Nil -> raise (Out_of_bounds n)
    | Cons (_, tail) -> drop (n - 1) tail
    | stream -> Delayed (function () -> drop (n - 1) (force stream))

let rec to_list = function stream ->
  match force stream with
    | Cons (head, tail) -> head :: to_list tail
    | Nil -> []
    | _ -> raise (Unexpected_pattern "StreamIII.to_list")

let rec exists f = function stream ->
  match force stream with
    | Nil -> false
    | Cons (head, tail) -> f head || exists f tail
    | _ -> raise (Unexpected_pattern "StreamIII.exits")

let rec for_all f = function stream ->
  match force stream with
    | Nil -> true
    | Cons (head, tail) -> f head && for_all f tail
    | _ -> raise (Unexpected_pattern "StreamIII.for_all")

let rec find f = function stream ->
  match force stream with
    | Nil -> raise Not_found
    | Cons (head, _) when f head -> head
    | Cons (_, tail) -> find f tail
    | _ -> raise (Unexpected_pattern "StreamIII.find")

let rec filter f = function
  | Nil -> Nil
  | Cons (head, tail) when f head -> Cons (head, filter f tail)
  | stream -> Delayed (function () -> filter f (force stream))
