open EdisonPervasives
  
type 
  'a stream = 'a streamCell ref and
  'a streamCell =
    | Nil
    | Cons of 'a * 'a stream
    | Delayed of (unit -> 'a stream)


(* L'un des principaux inconvnients de cette reprsentation est que l'on
 * a perdu le polymorphisme de la liste vide
 *
 *   # []
 *   - : 'a list
 *   # strict Nil
 *   - : 'a_ streamCell ref = { Pervasive.contents = Nil }
 *
 * On constate l'introduction d'une variable de type faible afin de prserver
 * la sret dy typage
 *)


(* force : 'a stream -> 'a stream *)

let force = function stream ->
  match !stream with
    | Delayed f -> let x = f () in stream := !x ; stream
    | other -> stream

let rec forceR_rec = function 
  | Delayed f -> forceR_rec (!(f ()))
  | other -> other

let forceR = function stream ->
  match !stream with
    | Delayed f -> stream := (forceR_rec !stream) ; stream
    | other -> stream

let lazy_ = function streamCell -> ref (Delayed (fun () -> ref streamCell))

let strict = function streamCell -> ref streamCell

let unit = function x -> strict (Cons (x, strict Nil))

let rec map f = function stream ->
  match !stream with
    | Nil -> strict Nil
    | Cons (head, tail) -> strict (Cons (f head, map f tail))
    | Delayed _ -> ref (Delayed (function () -> map f (force stream)))

let rec plus = fun a b ->
  match !a with
    | Nil -> b
    | Cons (head, tail) -> strict (Cons (head, plus tail b))
    | Delayed _ -> ref (Delayed (function () -> plus (force a) b))

let zero = function () -> strict Nil

let rec join = function stream ->
  match !stream with
    | Nil -> strict Nil
    | Cons (head, tail) -> plus head (join tail)
    | Delayed _ -> ref (Delayed (function () -> join (force stream)))

let bind = fun stream f -> join (map f stream)

let rec make = fun n x ->
  match n with
    | 0 -> strict Nil
    | _ -> lazy_ (Cons (x, make (n - 1) x))

let rec get = fun stream n ->
  match !stream with
    | Cons (x, _) when n = 0 -> x
    | Cons (_, tail) -> get tail (n - 1)
    | Delayed _ -> get (force stream) n
    |_ -> raise Empty

let rec set = fun stream n x ->
  match !stream with
    | Nil when n = 0 -> strict (Cons (x, strict Nil))
    | Cons (_, tail) when n = 0 -> strict (Cons (x, tail))
    | Cons (y, tail) -> strict (Cons (y, set tail (n - 1) x))
    | Delayed _ -> ref (Delayed (function () -> set (force stream) n x))
    | _ -> raise Empty

let rec from_list_rec = function
  | [] -> Nil
  | head :: tail -> Cons (head, strict (from_list_rec tail))

let from_list = function list -> strict (from_list_rec list)

let rec to_list = function stream ->
  match !stream with
    | Nil -> []
    | Cons (x, tail) -> x :: to_list tail
    | Delayed _ -> to_list (force stream)

let rec to_list_append a = function stream ->
  match !stream with
    | Nil -> a
    | Cons (x, tail) -> x :: to_list_append a tail
    | Delayed _ -> to_list_append a (force stream)

let rec reverse_onto result = function stream ->
  match !stream with
    | Nil -> result
    | Cons (head, tail) -> reverse_onto (strict (Cons (head, result))) tail
    | Delayed _ -> reverse_onto result (force stream)

let reverse_append = fun stream result -> reverse_onto result stream

let reverse = function stream -> reverse_append (strict Nil) stream

let rec append_reverse_rec result = fun a b ->
  match (!a, !b) with
    | (_, Delayed _) -> append_reverse_rec result a (force b)
    | (Nil, streamCell) -> reverse_onto result (strict streamCell)
    | (streamCell, Nil) -> plus (strict streamCell) result
    | (Delayed _, Cons (_, _)) -> append_reverse_rec result (force a) b
    | (Cons (x, xs), Cons (y, ys)) ->
	strict (Cons (x, append_reverse_rec (strict (Cons (y, result))) xs ys))

let rec append_reverse = fun a b -> append_reverse_rec (strict Nil) a b

let rec reverse_iter f = function stream ->
  match !stream with
    | Nil -> ()
    | Cons (head, tail) -> reverse_iter f tail ; f head
    | Delayed _ -> reverse_iter f (force stream)

let cons x = function stream -> strict (Cons (x, stream))

let rec snoc = fun stream x ->
  match !stream with
    | Nil -> strict (Cons (x, strict Nil))
    | Cons (head, tail) -> strict (Cons (head, snoc tail x))
    | Delayed _ -> ref (Delayed (function () -> snoc (force stream) x))

let init = fun n f ->
  let rec init_rec = function
    | k when k = n -> strict Nil
    | k -> ref (Delayed (function () -> strict (Cons (f k, init_rec (k - 1)))))
  in
    init_rec 0

let rec left_view = function stream ->
  match !stream with
    | Nil -> raise Empty
    | Cons (head, tail) -> (head, tail)
    | Delayed _ -> left_view (force stream)

let rec left_head = function stream ->
  match !stream with
    | Nil -> raise Empty
    | Cons (head, _) -> head
    | Delayed _ -> left_head (force stream)

let rec left_tail = function stream ->
  match !stream with
    | Nil -> raise Empty
    | Cons (_, tail) -> tail
    | Delayed _ -> left_tail (force stream)

let right_view = function stream ->
  match !(forceR stream) with
    | Cons (x, _) ->
	let right_head = ref x in
	let rec right_view_rec = function
	  | Cons (x, tail) -> 
	      (match !(forceR tail) with
		| Nil -> right_head := x ; strict Nil
		| other -> strict (Cons (x, right_view_rec other))
	      )
	  | _ -> raise (Unexpected_pattern "Stream.right_view")
	in
	let right_tail = right_view_rec !(forceR stream) in
	  (right_tail, !right_head)
    | _ -> raise Empty
	  
let rec right_tail_rec = function
  | Cons (head, tail) ->
      (match !(forceR tail) with
	 | Nil -> strict Nil
	 | streamCell -> strict (Cons (head, right_tail_rec streamCell))
      )	
  | _ -> raise Empty

let right_tail = function stream -> right_tail_rec !(forceR stream)

let rec right_head_rec = function
  | Cons (head, tail) ->
      (match !(forceR tail) with
	 | Nil -> head
	 | stream -> right_head_rec stream
      )
  | _ -> raise Empty
      
let right_head = function stream -> right_head_rec !(forceR stream)

let rec is_empty = function stream ->
  match !stream with
    | Nil -> true
    | Delayed _ -> is_empty (force stream)
    | _ -> false

let rec size = function stream ->
  match !stream with
    | Nil -> 0
    | Cons (_, tail) -> 1 + size tail
    | Delayed _ -> size (force stream)

let rec concat_map = fun f stream -> join (map f stream)

let rec fold_right f e = function stream ->
  match !stream with
    | Nil -> e
    | Cons (head, tail) -> f head (fold_right f e tail)
    | Delayed _ -> fold_right f e (force stream)

let rec fold_left f e = function stream ->
  match !stream with
    | Nil -> e
    | Cons (head, tail) -> fold_left f (f e head) tail
    | Delayed _ -> fold_left f e (force stream)

let rec iter f = function stream ->
  match !stream with
    | Nil -> ()
    | Cons (head, tail) -> f head ; iter f tail
    | Delayed _ -> iter f (force stream)

let rec reverse_iter f = function stream ->
  match !stream with
    | Nil -> ()
    | Cons (head, tail) -> reverse_iter f tail ; f head
    | Delayed _ -> reverse_iter f (force stream)

let take n = function stream ->
  let rec take_rec k = function stream ->
    if k = 0 then strict Nil
    else match !stream with
      | Nil -> raise (Out_of_bounds n)
      | Cons (head, tail) -> strict (Cons (head, take_rec (k - 1) tail))
      | Delayed _ -> ref (Delayed (function () -> take_rec k (force stream)))
  in
    if n < 0 then raise (Invalid_argument "take : argument ngatif")
    else take_rec n stream

let drop n = function stream ->
  let rec drop_rec k = function stream ->
    match !stream with
      | _ when k = 0 -> stream
      | Nil -> raise (Out_of_bounds n)
      | Cons (_, tail) -> drop_rec (k - 1) tail
      | Delayed _ -> ref (Delayed (function () -> drop_rec k (force stream)))
  in
    if n < 0 then raise (Invalid_argument "drop : argument ngatif")
    else drop_rec n stream

let split_at n = function stream ->
  let rec split_rec k left = function stream ->
    match !stream with
      | Nil when k = 0 -> (from_list (List.reverse left), stream)
      | Cons (head, tail) -> split_rec (k - 1) (head :: left) tail
      | Delayed _ -> split_rec k left (force stream)
      | _ -> raise (Out_of_bounds n)
  in
    split_rec n [] stream

let sub = fun starti endi stream ->
  let rec sub_rec n = function stream ->
    match !stream with
      | Nil when n = endi -> strict Nil
      | Nil -> raise (Out_of_bounds endi)
      | Cons (_, tail) when n < starti -> sub_rec (n + 1) tail
      | _ when n > endi -> strict Nil
      | Cons (head, tail) -> strict (Cons (head, sub_rec (n + 1) tail))
      | Delayed _ -> ref (Delayed (function () -> sub_rec n (force stream)))
  in
    if (starti < 0) then raise (Invalid_argument "sub : argument ngatif")
    else if (starti > endi) then raise (Invalid_argument "sub : start > end")
    else sub_rec 0 stream
      
let rec exists f = function stream ->
  match !stream with
    | Nil -> false
    | Cons (head, tail) -> f head || exists f tail
    | Delayed _ -> exists f (force stream)

let rec for_all f = function stream ->
  match !stream with
    | Nil -> true
    | Cons (head, tail) -> f head && for_all f tail
    | Delayed _ -> for_all f (force stream)
      
let rec find f = function stream ->
  match !stream with
    | Nil -> raise Not_found
    | Cons (head, _) when f head -> head
    | Cons (_, tail) -> find f tail
    | Delayed _ -> find f (force stream)

let rec filter f = function stream ->
  match !stream with
    | Nil -> strict Nil
    | Cons (head, tail) when f head -> strict (Cons (head, filter f tail))
    | Cons (_, tail) -> filter f tail
    | Delayed _ -> ref (Delayed (function () -> filter f (force stream)))

let rec partition f = function stream ->
  (filter f stream, filter (fun x -> not (f x)) stream)

let rec take_while f = function stream ->
  match !stream with
    | Cons (head, tail) when f head -> strict (Cons (head, take_while f tail))
    | Delayed _ -> ref (Delayed (function () -> take_while f (force stream)))
    | _ -> strict Nil

let rec drop_while f = function stream ->
  match !stream with
    | Cons (head, tail) when f head -> strict (Cons (head, drop_while f tail))
    | Delayed _ -> ref (Delayed (function () -> drop_while f (force stream)))
    | _ -> stream

let rec split_rec f left = function stream ->
  match !stream with
    | Cons (head, tail) when f head -> split_rec f (head :: left) tail
    | Delayed _ -> split_rec f left (force stream)
    | _ -> (from_list (List.reverse left), stream)

let split_while = fun f stream -> split_rec f [] stream


(** Index based operations **)


let rec mapi_rec f k = function stream ->
  match !stream with
    | Nil -> strict Nil
    | Cons (head, tail) -> strict (Cons (f k head, mapi_rec f (k + 1) tail))
    | Delayed _ -> ref (Delayed (function () -> mapi_rec f k (force stream)))

let map_with_index f = function stream -> mapi_rec f 0 stream

let rec iteri_rec f k = function stream ->
  match !stream with
    | Nil -> ()
    | Cons (head, tail) -> f k head ; iteri_rec f (k + 1) tail
    | Delayed _ -> iteri_rec f k (force stream)

let rec iter_with_index f = function stream -> iteri_rec f 0 stream

let rec findi_rec f k = function stream ->
  match !stream with
    | Nil -> raise Not_found
    | Cons (head, _) when f k head -> head
    | Cons (_, tail) -> findi_rec f (k + 1) tail
    | Delayed _ -> findi_rec f k (force stream)

let find_with_index f = function stream -> findi_rec f 0 stream

let rec filteri_rec f k = function stream ->
  match !stream with
    | Nil -> strict Nil
    | Cons (head, tail) when f k head -> 
	strict (Cons (head, filteri_rec f (k + 1) tail))
    | Cons (_, tail) -> filteri_rec f (k + 1) tail
    | Delayed _ -> 
	ref (Delayed (function () -> filteri_rec f k (force stream)))

let filter_with_index f = function stream -> filteri_rec f 0 stream

let rec foldri f e k = function stream ->
  match !stream with
    | Nil -> e
    | Cons (head, tail) -> f k head (foldri f e (k + 1) tail)
    | Delayed _ -> foldri f e k (force stream)

let fold_right_with_index f e = function stream -> foldri f e 0 stream

let rec foldli f e k = function stream ->
  match !stream with
    | Nil -> e
    | Cons (head, tail) -> foldli f (f k e head) (k + 1) tail
    | Delayed _ -> foldli f e k (force stream)

let fold_left_with_index f e = function stream -> foldli f e 0 stream

let rec zip = fun a b ->
  match (!(force a), !(force b)) with
    | (Nil, Nil) -> strict Nil
    | (Cons (hda, tla) , Cons (hdb, tlb)) -> 
	strict (Cons ((hda, hdb), zip tla tlb))
    | _ -> raise 
	(Invalid_argument "zip : les squences doivent tre de mme taille")

let rec unzip_left = function stream ->
  match !stream with
    | Nil -> strict Nil
    | Cons ((a, _), tail) -> strict (Cons (a, unzip_left tail))
    | Delayed _ -> unzip_left (force stream)

let rec unzip_right = function stream ->
  match !stream with
    | Nil -> strict Nil
    | Cons ((_, b), tail) -> strict (Cons (b, unzip_right tail))
    | Delayed _ -> unzip_right (force stream)

let unzip = function stream -> (unzip_left stream, unzip_right stream)

let empty = zero
let append = plus
	       
let blit = fun a b c d e -> raise (Not_working "blit")
let fill = fun a b c d -> raise (Not_working "fill")
  

module StreamSeq = struct

  type 
    'a streamCell =
    | Nil
    | Cons of 'a * 'a seq
    | Delayed of (unit -> 'a seq)
  and
    'a seq = 'a streamCell ref


  (* force : 'a stream -> 'a stream *)

  let force = function stream ->
    match !stream with
      | Delayed f -> let x = f () in stream := !x ; stream
      | other -> stream

  let rec forceR_rec = function 
    | Delayed f -> forceR_rec (!(f ()))
    | other -> other

  let forceR = function stream ->
    match !stream with
      | Delayed f -> stream := (forceR_rec !stream) ; stream
      | other -> stream

  let lazy_ = function streamCell -> ref (Delayed (fun () -> ref streamCell))

  let strict = function streamCell -> ref streamCell

  let unit = function x -> strict (Cons (x, strict Nil))

  let rec map f = function stream ->
    match !stream with
      | Nil -> strict Nil
      | Cons (head, tail) -> strict (Cons (f head, map f tail))
      | Delayed _ -> ref (Delayed (function () -> map f (force stream)))

  let rec plus = fun a b ->
    match !a with
      | Nil -> b
      | Cons (head, tail) -> strict (Cons (head, plus tail b))
      | Delayed _ -> ref (Delayed (function () -> plus (force a) b))

  let zero = function () -> strict Nil

  let rec join = function stream ->
    match !stream with
      | Nil -> strict Nil
      | Cons (head, tail) -> plus head (join tail)
      | Delayed _ -> ref (Delayed (function () -> join (force stream)))

  let bind = fun stream f -> join (map f stream)

  let rec make = fun n x ->
    match n with
      | 0 -> strict Nil
      | _ -> lazy_ (Cons (x, make (n - 1) x))

  let rec get = fun stream n ->
    match !stream with
      | Cons (x, _) when n = 0 -> x
      | Cons (_, tail) -> get tail (n - 1)
      | Delayed _ -> get (force stream) n
      |_ -> raise Empty

  let rec set = fun stream n x ->
    match !stream with
      | Nil when n = 0 -> strict (Cons (x, strict Nil))
      | Cons (_, tail) when n = 0 -> strict (Cons (x, tail))
      | Cons (y, tail) -> strict (Cons (y, set tail (n - 1) x))
      | Delayed _ -> ref (Delayed (function () -> set (force stream) n x))
      | _ -> raise Empty

  let rec from_list_rec = function
    | [] -> Nil
    | head :: tail -> Cons (head, strict (from_list_rec tail))

  let from_list = function list -> strict (from_list_rec list)

  let rec to_list = function stream ->
    match !stream with
      | Nil -> []
      | Cons (x, tail) -> x :: to_list tail
      | Delayed _ -> to_list (force stream)

  let rec to_list_append a = function stream ->
    match !stream with
      | Nil -> a
      | Cons (x, tail) -> x :: to_list_append a tail
      | Delayed _ -> to_list_append a (force stream)

  let rec reverse_onto result = function stream ->
    match !stream with
      | Nil -> result
      | Cons (head, tail) -> reverse_onto (strict (Cons (head, result))) tail
      | Delayed _ -> reverse_onto result (force stream)

  let reverse_append = fun stream result -> reverse_onto result stream

  let reverse = function stream -> reverse_append (strict Nil) stream

  let rec append_reverse_rec result = fun a b ->
    match (!a, !b) with
      | (_, Delayed _) -> append_reverse_rec result a (force b)
      | (Nil, streamCell) -> reverse_onto result (strict streamCell)
      | (streamCell, Nil) -> plus (strict streamCell) result
      | (Delayed _, Cons (_, _)) -> append_reverse_rec result (force a) b
      | (Cons (x, xs), Cons (y, ys)) ->
	  strict (Cons (x, append_reverse_rec (strict (Cons (y, result))) xs ys))

  let rec append_reverse = fun a b -> append_reverse_rec (strict Nil) a b

  let rec reverse_iter f = function stream ->
    match !stream with
      | Nil -> ()
      | Cons (head, tail) -> reverse_iter f tail ; f head
      | Delayed _ -> reverse_iter f (force stream)

  let rec cons x = function stream -> strict (Cons (x, stream))

  let rec snoc = fun stream x ->
    match !stream with
      | Nil -> strict (Cons (x, strict Nil))
      | Cons (head, tail) -> strict (Cons (head, snoc tail x))
      | Delayed _ -> ref (Delayed (function () -> snoc (force stream) x))

  let init = fun n f ->
    let rec init_rec = function
      | k when k = n -> strict Nil
      | k -> 
	  ref (Delayed (function () -> strict (Cons (f k, init_rec (k - 1)))))
    in
      init_rec 0

  let rec left_view = function stream ->
    match !stream with
      | Nil -> raise Empty
      | Cons (head, tail) -> (head, tail)
      | Delayed _ -> left_view (force stream)

  let rec left_head = function stream ->
    match !stream with
      | Nil -> raise Empty
      | Cons (head, _) -> head
      | Delayed _ -> left_head (force stream)

  let rec left_tail = function stream ->
    match !stream with
      | Nil -> raise Empty
      | Cons (_, tail) -> tail
      | Delayed _ -> left_tail (force stream)

  let right_view = function stream ->
    match !(forceR stream) with
      | Cons (x, _) ->
	  let right_head = ref x in
	  let rec right_view_rec = function
	    | Cons (x, tail) -> 
		(match !(forceR tail) with
		   | Nil -> right_head := x ; strict Nil
		   | other -> strict (Cons (x, right_view_rec other))
		)
	    | _ -> raise (Unexpected_pattern "Stream.right_view")
	  in
	  let right_tail = right_view_rec !(forceR stream) in
	    (right_tail, !right_head)
      | _ -> raise Empty
	  
  let rec right_tail_rec = function
    | Cons (head, tail) ->
	(match !(forceR tail) with
	   | Nil -> strict Nil
	   | streamCell -> strict (Cons (head, right_tail_rec streamCell))
	)	
    | _ -> raise Empty

  let right_tail = function stream -> right_tail_rec !(forceR stream)

  let rec right_head_rec = function
    | Cons (head, tail) ->
	(match !(forceR tail) with
	   | Nil -> head
	   | stream -> right_head_rec stream
	)
    | _ -> raise Empty
	
  let right_head = function stream -> right_head_rec !(forceR stream)

  let rec is_empty = function stream ->
    match !stream with
      | Nil -> true
      | Delayed _ -> is_empty (force stream)
      | _ -> false

  let rec size = function stream ->
    match !stream with
      | Nil -> 0
      | Cons (_, tail) -> 1 + size tail
      | Delayed _ -> size (force stream)

  let rec concat_map = fun f stream -> join (map f stream)

  let rec fold_right f e = function stream ->
    match !stream with
      | Nil -> e
      | Cons (head, tail) -> f head (fold_right f e tail)
      | Delayed _ -> fold_right f e (force stream)

  let rec fold_left f e = function stream ->
    match !stream with
      | Nil -> e
      | Cons (head, tail) -> fold_left f (f e head) tail
      | Delayed _ -> fold_left f e (force stream)

  let rec iter f = function stream ->
    match !stream with
      | Nil -> ()
      | Cons (head, tail) -> f head ; iter f tail
      | Delayed _ -> iter f (force stream)

  let rec reverse_iter f = function stream ->
    match !stream with
      | Nil -> ()
      | Cons (head, tail) -> reverse_iter f tail ; f head
      | Delayed _ -> reverse_iter f (force stream)

  let take n = function stream ->
    let rec take_rec k = function stream ->
      if k = 0 then strict Nil
      else match !stream with
	| Nil -> raise (Out_of_bounds n)
	| Cons (head, tail) -> strict (Cons (head, take_rec (k - 1) tail))
	| Delayed _ -> ref (Delayed (function () -> take_rec k (force stream)))
    in
      if n < 0 then raise (Invalid_argument "take : argument ngatif")
      else take_rec n stream

  let drop n = function stream ->
    let rec drop_rec k = function stream ->
      match !stream with
	| _ when k = 0 -> stream
	| Nil -> raise (Out_of_bounds n)
	| Cons (_, tail) -> drop_rec (k - 1) tail
	| Delayed _ -> ref (Delayed (function () -> drop_rec k (force stream)))
    in
      if n < 0 then raise (Invalid_argument "drop : argument ngatif")
      else drop_rec n stream

  let split_at n = function stream ->
    let rec split_rec k left = function stream ->
      match !stream with
	| Nil when k = 0 -> (from_list (List.reverse left), stream)
	| Cons (head, tail) -> split_rec (k - 1) (head :: left) tail
	| Delayed _ -> split_rec k left (force stream)
	| _ -> raise (Out_of_bounds n)
    in
      split_rec n [] stream

  let sub = fun starti endi stream ->
    let rec sub_rec n = function stream ->
      match !stream with
	| Nil when n = endi -> strict Nil
	| Nil -> raise (Out_of_bounds endi)
	| Cons (_, tail) when n < starti -> sub_rec (n + 1) tail
	| _ when n > endi -> strict Nil
	| Cons (head, tail) -> strict (Cons (head, sub_rec (n + 1) tail))
	| Delayed _ -> ref (Delayed (function () -> sub_rec n (force stream)))
    in
      if (starti < 0) then raise (Invalid_argument "sub : argument ngatif")
      else if (starti > endi) then raise (Invalid_argument "sub : start > end")
      else sub_rec 0 stream
	
  let rec exists f = function stream ->
    match !stream with
      | Nil -> false
      | Cons (head, tail) -> f head || exists f tail
      | Delayed _ -> exists f (force stream)

  let rec for_all f = function stream ->
    match !stream with
      | Nil -> true
      | Cons (head, tail) -> f head && for_all f tail
      | Delayed _ -> for_all f (force stream)
	  
  let rec find f = function stream ->
    match !stream with
      | Nil -> raise Not_found
      | Cons (head, _) when f head -> head
      | Cons (_, tail) -> find f tail
      | Delayed _ -> find f (force stream)

  let rec filter f = function stream ->
    match !stream with
      | Nil -> strict Nil
      | Cons (head, tail) when f head -> strict (Cons (head, filter f tail))
      | Cons (_, tail) -> filter f tail
      | Delayed _ -> ref (Delayed (function () -> filter f (force stream)))

  let rec partition f = function stream ->
    (filter f stream, filter (fun x -> not (f x)) stream)

  let rec take_while f = function stream ->
    match !stream with
      | Cons (head, tail) when f head -> 
	  strict (Cons (head, take_while f tail))
      | Delayed _ -> ref (Delayed (function () -> take_while f (force stream)))
      | _ -> strict Nil

  let rec drop_while f = function stream ->
    match !stream with
      | Cons (head, tail) when f head -> 
	  strict (Cons (head, drop_while f tail))
      | Delayed _ -> ref (Delayed (function () -> drop_while f (force stream)))
      | _ -> stream

  let rec split_rec f left = function stream ->
    match !stream with
      | Cons (head, tail) when f head -> split_rec f (head :: left) tail
      | Delayed _ -> split_rec f left (force stream)
      | _ -> (from_list (List.reverse left), stream)

  let split_while = fun f stream -> split_rec f [] stream


  (** Index based operations **)


  let rec mapi_rec f k = function stream ->
    match !stream with
      | Nil -> strict Nil
      | Cons (head, tail) -> strict (Cons (f k head, mapi_rec f (k + 1) tail))
      | Delayed _ -> ref (Delayed (function () -> mapi_rec f k (force stream)))

  let map_with_index f = function stream -> mapi_rec f 0 stream

  let rec iteri_rec f k = function stream ->
    match !stream with
      | Nil -> ()
      | Cons (head, tail) -> f k head ; iteri_rec f (k + 1) tail
      | Delayed _ -> iteri_rec f k (force stream)

  let rec iter_with_index f = function stream -> iteri_rec f 0 stream

  let rec findi_rec f k = function stream ->
    match !stream with
      | Nil -> raise Not_found
      | Cons (head, _) when f k head -> head
      | Cons (_, tail) -> findi_rec f (k + 1) tail
      | Delayed _ -> findi_rec f k (force stream)

  let find_with_index f = function stream -> findi_rec f 0 stream

  let rec filteri_rec f k = function stream ->
    match !stream with
      | Nil -> strict Nil
      | Cons (head, tail) when f k head -> 
	  strict (Cons (head, filteri_rec f (k + 1) tail))
      | Cons (_, tail) -> filteri_rec f (k + 1) tail
      | Delayed _ -> 
	  ref (Delayed (function () -> filteri_rec f k (force stream)))

  let filter_with_index f = function stream -> filteri_rec f 0 stream

  let rec foldri f e k = function stream ->
    match !stream with
      | Nil -> e
      | Cons (head, tail) -> f k head (foldri f e (k + 1) tail)
      | Delayed _ -> foldri f e k (force stream)

  let fold_right_with_index f e = function stream -> foldri f e 0 stream

  let rec foldli f e k = function stream ->
    match !stream with
      | Nil -> e
      | Cons (head, tail) -> foldli f (f k e head) (k + 1) tail
      | Delayed _ -> foldli f e k (force stream)

  let fold_left_with_index f e = function stream -> foldli f e 0 stream

  let rec zip = fun a b ->
    match (!(force a), !(force b)) with
      | (Nil, Nil) -> strict Nil
      | (Cons (hda, tla) , Cons (hdb, tlb)) -> 
	  strict (Cons ((hda, hdb), zip tla tlb))
      | _ -> raise 
	  (Invalid_argument "zip : les squences doivent tre de mme taille")

  let rec unzip_left = function stream ->
    match !stream with
      | Nil -> strict Nil
      | Cons ((a, _), tail) -> strict (Cons (a, unzip_left tail))
      | Delayed _ -> unzip_left (force stream)

  let rec unzip_right = function stream ->
    match !stream with
      | Nil -> strict Nil
      | Cons ((_, b), tail) -> strict (Cons (b, unzip_right tail))
      | Delayed _ -> unzip_right (force stream)

  let unzip = function stream -> (unzip_left stream, unzip_right stream)

  let empty = zero
  let append = plus
		 
  let blit = fun a b c d e -> raise (Not_working "blit")
  let fill = fun a b c d -> raise (Not_working "fill")

end
