open EdisonPervasives
  
type
  'a stream = { mutable data : 'a streamCell }  and
  'a streamCell =
    | Nil
    | Cons of 'a * 'a stream
    | Append of 'a stream * 'a stream
    | Delayed of (unit -> 'a stream)
    | Generator of int * (int -> int) * (int -> 'a)

(* Nous introduisons  prsent un oprateur de concatnation et un
 * gnrateur de valeurs indices :
 *
 *  Generator (index, succ, gen)
 *
 * Le calcul de la valeur suivante se fait par gen (succ (index))
 * laissant le gnrateur dans l'tat (succ (index), succ, gen)
 * Cela permet d'offrir des fonctions [drop] et [take] relativement
 * efficaces
 * 
 * - la fonction [gen] renvoie une valeur de type ['a] ou bien lve l'exception
 * [Empty] pour signaler la fin du flot 
 * - la fonction [succ] renvoie une valeur de type [int] ou bien lve 
 * l'exception [Empty] pour signaler la fin du flot
 *)

(** Utils **)

let makeStream = function x -> { data = x }

let rec last_element = function stream ->
  match stream.data with
    | Cons (_, tail) -> last_element tail
    | Append (a, b) -> last_element a
    | Delayed _ | Generator _ | Nil -> stream

let rec force = function stream ->
  (match stream.data with
    | Delayed f -> let x = f () in stream.data <- x.data
    | Generator (index, succ, gen) ->
	(try
	   (let next_index = succ index in
	      stream.data <-
	      Cons (
		gen (next_index), 
		makeStream (Generator (next_index, succ, gen))))
	 with Empty -> stream.data<- Nil)
    | Append (a, b) ->
	let new_stream = force a in
	let last = last_element new_stream in
	  stream.data <- new_stream.data ;
	  last.data <- Append (makeStream last.data, b)
    | Cons _ | Nil -> ()
  ) ;
  stream

let o = fun f g -> (function x -> f (g x))

(** Monadic operations **)

let unit = function x -> makeStream (Cons (x, makeStream Nil))

let rec map f = function stream ->
  match stream.data with
    | Nil -> makeStream Nil
    | Cons (head, tail) -> makeStream (Cons (f head, map f tail))
    | Append (a, b) -> makeStream (Append (map f a, map f b))
    | Delayed _ -> makeStream (Delayed (function () -> map f (force stream)))
    | Generator (i, s, g) -> makeStream (Generator (i, s, o f g))

let rec plus = fun a b ->
  match a.data with
    | Nil -> b
    | Cons (head, tail) -> makeStream (Cons (head, plus tail b))
    | Append (c, d) -> plus (plus c d) b
    | Delayed _ -> makeStream (Delayed (function () -> plus (force a) b))
    | Generator (i, s, g) -> makeStream (Append (a, b))

let zero = function () -> makeStream Nil

let rec make = fun n x ->
  match n with
    | 0 -> makeStream Nil
    | _ -> makeStream (Cons (x, make (n - 1) x))

let rec get = fun stream n ->
  match stream.data with
    | Nil -> raise Empty
    | Cons (x, _) when n = 0 -> x
    | Cons (_, a) | Append (a, _) -> get a (n - 1)
    | Delayed _ | Generator _  -> get (force stream) n

let rec set = fun stream n x ->
  match stream.data with
    | Cons (_, tail) when n = 0 -> makeStream (Cons (x, tail))
    | Cons (head, tail) -> makeStream (Cons (head, set tail (n - 1) x))
    | Append (a, _) -> set a n x
    | Delayed _ -> makeStream (Delayed (function () -> set (force stream) n x))
    | Generator _ -> set (force stream) n x
    | Nil when n = 0 -> makeStream (Cons (x, makeStream Nil))
    | Nil -> raise Empty

let rec from_list = function
  | [] -> makeStream Nil
  | head :: tail -> makeStream (Cons (head, from_list tail))

let from_function = function f -> makeStream (Generator (0, succ, f))

let from_generator = fun i s f -> makeStream (Generator (i, s, f))

let rec to_list = function stream ->
  match stream.data with
    | Nil -> []
    | Cons (head, tail) -> head :: to_list tail
    | Append (a, b) -> to_list a @ to_list b
    | Delayed _ | Generator _ -> to_list (force stream)

let cons x = function stream -> makeStream (Cons (x, stream))

let rec left_view = function stream ->
  match stream.data with
    | Nil -> raise Empty
    | Cons (head, tail) -> (head, tail)
    | Append (a, b) -> let (ha, ta) = left_view a in (ha, plus ta b)
    | Delayed _ | Generator _ -> left_view (force stream)

let rec left_head = function stream ->
  match stream.data with
    | Nil -> raise Empty
    | Cons (head, _) -> head
    | Append (a, _) -> left_head a
    | Delayed _ | Generator _ -> left_head (force stream)

let rec left_tail = function stream ->
  match stream.data with
    | Nil -> raise Empty
    | Cons (_, tail) -> tail
    | Append (a, b) -> plus (left_tail a) b
    | Delayed _ | Generator _ -> left_tail (force stream)

let rec is_empty = function stream ->
  match stream.data with
    | Nil -> true
    | Cons _ -> false
    | Append (a, _) -> is_empty a
    | Delayed _ | Generator _ -> is_empty (force stream)

let rec size = function stream ->
  match stream.data with
    | Nil -> 0
    | Cons (_, tail) -> 1 + size tail
    | Append (a, b) -> size a + size b
    | Delayed _ | Generator _ -> size (force stream)

let rec fold_right f e = function stream ->
  match stream.data with
    | Nil -> e
    | Cons (head, tail) -> f head (fold_right f e tail)      
    | Append (a, b) -> fold_right f (fold_right f e b) a 
    | Delayed _ | Generator _ -> fold_right f e (force stream)

let rec fold_left f e = function stream ->
  match stream.data with
    | Nil -> e
    | Cons (head, tail) -> fold_left f (f e head) tail
    | Append (a, b) -> fold_left f (fold_left f e a) b
    | Delayed _ | Generator _ -> fold_left f e (force stream)

let rec iter f = function stream ->
  match stream.data with
    | Nil -> ()
    | Cons (head, tail) -> f head ; iter f tail
    | Append (a, b) -> iter f a ; iter f b
    | Delayed _ | Generator _ -> iter f (force stream)

let take n = function stream ->
  let rec take_rec k = function stream ->
    if k = 0 then makeStream Nil
    else match stream.data with
      | Nil -> raise (Out_of_bounds n)
      | Cons (head, tail) -> makeStream (Cons (head, take_rec (k - 1) tail))
      | Append (a, b) -> take_rec k (plus a b)
      | Delayed _ ->
	  makeStream (Delayed (function () -> take_rec k (force stream)))
      | Generator (i, s, f) ->
	  let count = ref k in
	  let s' = function index ->
	    if !count = 0 then raise Empty
	    else decr count ; s index
	  in
	    makeStream (Generator (i, s', f))
  in
    if n < 0 then raise (Invalid_argument "StreamVI.take : argument ngatif")
    else take_rec n stream

let rec power f = function
  | 1 -> f
  | n -> function x -> f (power f (n - 1) x)

let drop n = function stream ->
  let rec drop_rec k = function stream ->
    match stream.data with
      | _ when k = 0 -> stream
      | Nil -> raise (Out_of_bounds n)
      | Cons (_, tail) -> drop_rec (k - 1) tail
      | Append (a, b) -> drop_rec k (plus a b)
      | Delayed _ -> 
	  makeStream (Delayed (function () -> drop_rec k (force stream)))
      | Generator (i, s, f) -> let succ = power s k in 
	  makeStream (Generator (succ i, s, f))
  in
    drop_rec n stream
